# -*- coding: utf-8 -*-
# junky lib: dataset.BertDataset
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides implementation of torch.utils.data.Dataset for word-level input.
"""
import logging
import sys
#logging.basicConfig(level=logging.ERROR)
#if not sys.warnoptions:
#    import warnings
#    warnings.simplefilter('ignore')
#    os.environ['PYTHONWARNINGS'] = 'ignore'
from junky import CPU, absmax_torch, pad_array_torch, \
                  pad_sequences_with_tensor
from junky.dataset.base_dataset import BaseDataset
from time import time
import torch
from torch import Tensor, tensor
from tqdm import tqdm

# to suppress transformers' warnings
#logging.getLogger('transformers.tokenization_utils_base').setLevel(logging.ERROR)
#logging.getLogger('pytorch_pretrained_bert.tokenization').setLevel(logging.ERROR)
logging.getLogger('transformers').setLevel(logging.ERROR)


class BertDataset(BaseDataset):
    """
    `torch.utils.data.Dataset` for word-level input with contextual
    embeddings.

    Args:
        model: one of the token classification models from the `transformers`
            package. It should be created with config containing
            output_hidden_states=True. NB: Don't forget to set model in the
            eval mode before use it with this class.
        tokenizer: the tokenizer from `transformers` package corresponding to
            `model` chosen.
        int_tensor_dtype: dtype for int tensors: torch.dtype.
        sentences: sequences of words: list([list([str])]). If not ``None``,
            they will be transformed and saved. NB: All the sentences must
            not be empty.
        All other args are params for the `.transform()` method. They are used
            only if *sentences* is not ``None``. You can use any args but
            `save` that is set to `True`.
    """
    overlap_shift = .5  # Defines the overlap's `shift` from the sentence's
    # start. We count it in words, so, if sentence has `9` words, the shift
    # will be `int(.5 * 9)`, i.e., `4`. The minimum value for `shift` is `1`.
    # If you set `ds.overlap_shift` > `1`, we will treat it as absolute value
    # (but reduce it to `max_len` if your `ds.overlap_shift` would be greater.
    overlap_border = 2  # The overlap is processed as follows. The left zone
    # of width equal to `ds.overlap_border` is taken from the earlier part of
    # the sentence; the right zone - from the later. The zone between borders
    # is calculated as weighted sum of both parts. The weights are
    # proportional to the distance to the middle of the zone: the beginning
    # has dominance to the left from the middle, the ending has dominance to
    # the right. In the very middle (if it exists), both weights are equal to
    # `.5`. If you set `ds.overlap_border` high enough (greater than
    # `(max_len - shift) / 2`) or `None`, it will be set to the middle of the
    # overlap zone. Thus, weighted algorithm will be dwindled.  Also note that
    # weights are applied to tokens (not subtokens). I.e. all subtokens of any
    # particular token have equal weights when summing.
    use_batch_max_len = True  # Either we want to use the length of the
    # longest sentence in the batch instead of the `max_len` param of
    # `.transform()`. We use it only if that length is less than `max_len`,
    # and as result, with high *max_len*, we have a substantial speed increase
    # without any quality change or resulting data.
    sort_dataset = True  # Do we want to sort the dataset before feeding it to
    # `ds.model`. With high **max_len** it highly increases processing speed,
    # and affects resulting data only because of different sentences' grouping
    # (deviation is about `1e-7`).

    @property
    def vec_size(self):
        if self.data:
            if isinstance(self.data[0], list):
                res = self.data[0][0].shape[-1]
            else:
                res = self.data[0].shape[-1]
        else:
            res = self.model.config.hidden_size
        return res

    def __init__(self, model, tokenizer, int_tensor_dtype=torch.int64,
                 sentences=None, **kwargs):
        assert model.config.output_hidden_states, \
               'ERROR: BERT model was created with ' \
               '`output_hidden_states=False` option. To use it in ' \
               'BertDataset, the model must be created with ' \
               '`output_hidden_states=True`'
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.int_tensor_dtype = int_tensor_dtype
        if sentences:
            self.transform(sentences, save=True, **kwargs)

    def _pull_xtrn(self):
        xtrn = self.model, self.tokenizer
        self.model, self.tokenizer = None, None
        return xtrn

    def _push_xtrn(self, xtrn):
        self.model, self.tokenizer = xtrn

    @staticmethod
    def _aggregate_hidden_states(hidden_states, layer_ids=None,
                                 aggregate_op='mean'):
        if isinstance(layer_ids, int):
            hiddens = hidden_states[layer_ids]

        else:
            if layer_ids is None:
                hiddens = hidden_states
            else:
                hiddens = [hidden_states[x] for x in layer_ids]

            if aggregate_op == 'absmax':
                hiddens = absmax_torch(hiddens, dim=0)
            elif aggregate_op == 'cat':
                hiddens = torch.cat(hiddens, dim=-1)
            elif aggregate_op == 'first':
                hiddens = hiddens[0]
            elif aggregate_op == 'last':
                hiddens = hiddens[-1]
            elif aggregate_op == 'max':
                hiddens = torch.max(hiddens, dim=0)[0]
            elif aggregate_op == 'mean':
                hiddens = torch.mean(hiddens if isinstance(hiddens, Tensor) else
                                     torch.stack(hiddens, dim=0),
                                     dim=0)
            elif aggregate_op == 'sum':
                hiddens = torch.sum(hiddens if isinstance(hiddens, Tensor) else
                                     torch.stack(hiddens, dim=0),
                                     dim=0)
            else:
                RuntimeError(
                    'ERROR: unknown aggregate_op '
                    "(choose one of ['absmax', 'cat', 'max', 'mean', 'sum'])"
                )
        return hiddens

    def transform(self, sentences, max_len=None, batch_size=64,
                  hidden_ids=0, aggregate_hiddens_op='mean',
                  aggregate_subtokens_op='absmax', to=None,
                  with_grad=False, save=True, append=False, loglevel=1):
        """Convert *sentences* of words to the sequences of the corresponding
        contextual vectors and adjust their format for Dataset.

        *max_len* is a param for tokenizer. We'll transform lines of any
            length, but the quality is higher if *max_len* is greater.
            ``None`` (default) or `0` means the maximum for the model
            (usually, `512`).

        *batch_size* affects only on the execution time. Greater is faster,
            but big *batch_size* may be cause of CUDA Memory Error. If
            ``None`` or `0`, we'll try to convert all *sentences* with one
            batch.

        *hidden_ids*: hidden score layers that we need to aggregate. Allowed
            int or tuple of ints. If ``None``, we'll aggregate all the layers.

        *aggregate_hidden_op*: how to aggregate hidden scores. The ops
            allowed: 'absmax', 'cat', 'max', 'mean', 'sum'. For the 'absmax'
            method we take into account absolute values of the compared items.

        *aggregate_subtokens_op*: how to aggregate subtokens vectors to form
            only one vector for each input token. The ops allowed: ``None``,
            'absmax', 'append', 'first', 'last', 'max', 'mean', 'sum'. For the
            'absmax' method we take into account absolute values of the
            compared items.

        If you want to get the result placed on some exact device, specify the
        device with *to* param. If *to* is ``None`` (defautl), data will be
        placed to the very device that `ds.model` is used.

        *with_grad*: calculate gradients during forward propagation through
        self.model. Default is ``False``.

        If *save* is ``True``, we'll keep the converted sentences as the
        `Dataset` source.

        If *append* is ``True``, we'll append the converted sentences to the
        existing Dataset source. Elsewise (default), the existing Dataset
        source will be replaced. The param is used only if *save* is ``True``.

        *loglevel* can be set to `0`, `1` or `2`. `0` means no output.

        The result is depend on *aggregate_subtokens_op* param. If it is
        ``None``, then for each token we keeps in the result a tensor with
        stacked vectors for all its subtokens. Otherwise, if any
        *aggregate_subtokens_op* is used, each sentence will be converted to
        exactly one tensor of shape [<sentence length>, <vector size>]."""

        tokenizer_max_len = self.tokenizer.max_len \
                                if hasattr(self.tokenizer, 'max_len') else \
                            self.tokenizer.model_max_len \
                                if hasattr(self.tokenizer,
                                           'model_max_len') else \
                            512

        if not max_len:
            max_len = tokenizer_max_len
        assert max_len >= 16, 'ERROR: max len must be >= 16'
        assert max_len <= tokenizer_max_len, \
               'ERROR: max len must be <= {}'.format(tokenizer_max_len)
        valid_ops = ['absmax', 'cat', 'max', 'mean', 'sum']
        assert aggregate_hiddens_op in valid_ops, \
               'ERROR: unknown aggregate_hidden_op (choose one of {})' \
                   .format(valid_ops)
        valid_ops = [None, 'absmax', 'append', 'first', 'last', 'max', 'mean',
                     'sum']
        assert aggregate_subtokens_op in valid_ops, \
               'ERROR: unknown aggregate_subtokens_op (choose one of {})' \
                   .format(valid_ops)
        if self.data and append:
            assert (
                isinstance(self.data[0], list)
            and aggregate_subtokens_op is None
            ) or (
                isinstance(self.data[0], Tensor)
            and aggregate_subtokens_op is not None
            ), "ERROR: can't append data created with inconsistent " \
               'aggregate_subtokens_op'
        device = next(self.model.parameters()).device
        max_len_ = max_len - 2  # for [CLS] and [SEP]

        overlap_border = max_len if self.overlap_border is None else \
                         0 if self.overlap_border < 0 else \
                         self.overlap_border
        shift = int(max(min(max_len_, self.overlap_shift)
                        if self.overlap_shift >= 1 else
                    (max_len_ * self.overlap_shift), 1))
        # how many subtokens should be overlapped, e.g.:
        # max_len = 100, shift = 50, array = [0, 1, 2, ..., 120]
        # then array1 = [0, 1, 2, ..., 99], array2 = [50, 51, 52, ..., 120]
        # for every sentence it is rounded to tokens' borders (including)

        _src = sentences
        if loglevel >= 2:
            print('Tokenizing')
            _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)
        time0 = time()
        # tokenize each token separately by the BERT tokenizer
        tokenized_sentences = \
            [[x for x in self.tokenizer.tokenize(x) or ['[UNK]']]
                 for x in _src] \
                if sentences and isinstance(sentences[0], str) else \
            [[self.tokenizer.tokenize(x) or ['[UNK]'] for x in x] \
                 for x in _src]
        if time() - time0 < 2 and loglevel == 1:
            loglevel = 0

        # number of subtokens in tokens of tokenized_sentences
        num_subtokens = [[len(x) for x in x] for x in tokenized_sentences]
        # for each subtoken we keep index of its token
        sub_to_kens = [[
            x for x in [[i] * x for i, x in enumerate(x)]
              for x in x
        ] for x in num_subtokens]  # [[0, 0, 0, 1, 2, 3, 3, 4, ...], ...]
        # for each token we keep its start in flatten tokenized_sentences
        token_starts = []  # [[0, 3, 4, 5, 7, ...], ...]
        for sent in sub_to_kens:
            starts, prev_idx = [], -1
            for i, idx in enumerate(sent):
                if idx != prev_idx:
                    starts.append(i)
                    prev_idx = idx
            token_starts.append(starts)

        # flattening tokenized_sentences
        tokenized_sentences = [[x for x in x for x in x]
                                   for x in tokenized_sentences]
        # lengths of flattened tokenized_sentences
        sent_lens = [len(x) for x in tokenized_sentences]

        need_sort_dataset = self.sort_dataset \
                        and self.use_batch_max_len \
                        and len(tokenized_sentences) > batch_size
        ## sort tokenized sentences by lenght
        ######
        if need_sort_dataset:
            sent_lens, sorted_sent_ids = zip(*sorted(
                [(x, i) for i, x in enumerate(sent_lens)],
                reverse=True
            ))
            sent_lens = list(sent_lens)
            tokenized_sentences = [
                tokenized_sentences[i] for i in sorted_sent_ids
            ]
            sub_to_kens = [
                sub_to_kens[i] for i in sorted_sent_ids
            ]
            token_starts = [
                token_starts[i] for i in sorted_sent_ids
            ]
            ######

        def process_long_sentences(sents, sent_lens, sub_to_kens,
                                   token_starts, first_sent_idx):
            overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
            overlap_token_starts, overmap = [], [], [], [], []
            for sent_idx, (sent, sent_len, sub_token_ids, token_sub_ids) \
                    in enumerate(zip(sents, sent_lens, sub_to_kens,
                                     token_starts)):
                if sent_len > max_len_:
                    # индекс первого сабтокена в начальном слове фрагмента
                    # и индекс токена для первого сабтокена во фрагменте
                    sub0_token, token0_sub = \
                        sub_token_ids[0], token_sub_ids[0]
                    # индекс начального токена для области пересечения
                    over_tok = sub_token_ids[max_len_ - shift]
                    # вычитаем токен нулевого сабтокена, получаем индекс
                    # токена относительно текущего начала (если фразу уже
                    # делили, то это будет не 0)
                    over_tok_ = over_tok - sub0_token
                    # если в результате получаем 0, то переносим на
                    # следующий токен
                    if over_tok_ == 0:
                        over_tok += 1
                        over_tok_ = 1
                    # находим индекс сабтокена: из индекса сабтокена
                    # найденного токена вычитаем индекс текущего нулевого
                    # сабтокена
                    over_sub_ = token_sub_ids[over_tok_] - token0_sub
                    if over_sub_ > max_len_:
                        # it can appear only when over_tok_ == 1
                        raise RuntimeError(
                            ('ERROR: too long token (longer than '
                            f'effective max_len):\n{sent[:over_sub_]}')
                        )
                    overlap_sents.append(sent[over_sub_:])
                    overlap_sent_lens.append(sent_len - over_sub_)
                    overlap_sub_to_kens.append(sub_token_ids[over_sub_:])
                    overlap_token_starts.append(token_sub_ids[over_tok_:])
                    # индекс начального токена для отрезаемого фрагмента
                    cut_tok = sub_token_ids[max_len_]
                    # относительный индекс начального токена для отрезаемого
                    # фрагмента
                    cut_tok_ = cut_tok - sub0_token
                    cut_sub_ = token_sub_ids[cut_tok_] - token0_sub
                    sent[cut_sub_:] = []  # cut long sentence
                    overmap.append((first_sent_idx + sent_idx,
                                    over_tok, cut_tok))
            return overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
                   overlap_token_starts, overmap

        overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
        overlap_token_starts, overmap = process_long_sentences(
            tokenized_sentences, sent_lens, sub_to_kens,
            token_starts, 0
        )
        num_sents = len(tokenized_sentences)
        overmap = {num_sents + i: x for i, x in enumerate(overmap)}
        # overlap_sents: cutted parts of tokenized_sentences
        # overlap_sent_lens: their lens in subtokens
        # overlap_sub_to_kens: [sub -> token] map, indexes are absolute!
        # overlap_token_starts: [token_start -> first sub] map, absolute!
        # overmap: {sent_no: (orig_sent_no, orig_start_token, orig_end_token)}
        #     (orig_end_token is not included)

        while overlap_sents:
            tokenized_sentences += overlap_sents
            sent_lens += overlap_sent_lens
            sub_to_kens += overlap_sub_to_kens
            token_starts += overlap_token_starts
            overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
            overlap_token_starts, overmap_ = process_long_sentences(
                overlap_sents, overlap_sent_lens, overlap_sub_to_kens,
                overlap_token_starts, num_sents
            )
            num_sents = len(tokenized_sentences)
            for sent_idx, (orig_sent_idx, over_tok, cut_tok) in enumerate(overmap_):
                orig_sent_idx_ = overmap.get(orig_sent_idx)
                if orig_sent_idx_:
                    orig_sent_idx = orig_sent_idx_[0]
                overmap[num_sents + sent_idx] = \
                    (orig_sent_idx, over_tok, cut_tok)

        splitted_sent_lens = [len(x) for x in tokenized_sentences]

        if not batch_size:
            batch_size = num_sents

        data = []
        _src = range(0, num_sents, batch_size)
        if loglevel:
            print('Vectorizing')
            _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)

        ####### workaround for transformers v.4
        test_enc = self.tokenizer.encode_plus(
            text='', add_special_tokens=True, max_length=3,
            padding='max_length', return_tensors=None,
            return_token_type_ids=False, return_attention_mask=True
        )
        cls_id, sep_id, pad_id = test_enc['input_ids']
        att_mask1, _, att_mask0 = test_enc['attention_mask']
        #######

        for batch_i in _src:
            ####### workaround for transformers v.4
            '''
            batch_max_len = min(
                max(splitted_sent_lens[batch_i:batch_i + batch_size]) + 2,
                max_len
            ) if self.use_batch_max_len else max_len

            encoded_sentences = [
                self.tokenizer.encode_plus(text=sent,
                                           add_special_tokens=True,
                                           max_length=batch_max_len,
                                           #truncation=True,
                                           #pad_to_max_length=True,
                                           padding='longest',
                                           return_tensors='pt',
                                           return_attention_mask=True,
                                           return_overflowing_tokens=False)
                    for sent in tokenized_sentences[batch_i:batch_i
                                                          + batch_size]
            ]
            encoded_sentences = []
            for sent in tokenized_sentences[batch_i:batch_i + batch_size]:
                try:
                    self.tokenizer.encode_plus(text=sent,
                                               add_special_tokens=True,
                                               max_length=batch_max_len,
                                               #truncation=True,
                                               #pad_to_max_length=True,
                                               padding='longest',
                                               return_tensors='pt',
                                               return_attention_mask=True,
                                               return_overflowing_tokens=False)
                except TypeError as e:
                    print('batch_max_len =', batch_max_len)
                    print('type(sent) =', type(sent))
                    print('sent = [{}]'.format(sent))
                    raise e'''

            batch_max_len = min(
                max(splitted_sent_lens[batch_i:batch_i + batch_size]),
                max_len - 2
            ) if self.use_batch_max_len else max_len - 2

            input_ids = tensor(
                [([cls_id]
                + self.tokenizer.convert_tokens_to_ids(sent)
                + [sep_id] + [pad_id] * (batch_max_len - len(sent)))
                     for sent in tokenized_sentences[batch_i:batch_i
                                                           + batch_size]],
                dtype=self.int_tensor_dtype
            )
            attention_mask = (input_ids != pad_id).type(self.int_tensor_dtype)
            '''
            input_ids, attention_masks = zip(*[
                (x['input_ids'], x['attention_mask'])
                    for x in encoded_sentences
            ])'''
            #######

            with torch.set_grad_enabled(with_grad):
                hiddens = self.model(
                    #torch.cat(input_ids, dim=0).to(device),
                    input_ids.to(device),
                    token_type_ids=None,
                    #attention_mask=torch.cat(attention_masks, dim=0)
                    #                    .to(device)
                    attention_mask=attention_mask.to(device)
                )[-1]

                hiddens = self._aggregate_hidden_states(
                    hiddens, layer_ids=hidden_ids,
                    aggregate_op=aggregate_hiddens_op
                )

                if to:
                    hiddens = hiddens.to(to)
                data_device = hiddens.device

                # sub_to_kens: [sub -> token] map, indexes are absolute!
                # token_starts: [token -> first sub] map, absolute!
                # overmap: {sent_no: (orig_sent_no, orig_start_token,
                #                     orig_end_token)}
                #     (orig_end_token is not included)
                for sent_idx, sent in enumerate(hiddens, start=batch_i):
                    sent_fin = 1 + splitted_sent_lens[sent_idx]
                    if sent_idx in overmap:
                        orig_sent_idx, over_tok, cut_tok = overmap[sent_idx]
                        orig_data = data[orig_sent_idx]
                        overlap = cut_tok - over_tok  # overlap is in tokens
                        # overlap_border: how many tokens on the rims we don't
                        # touch
                        if overlap > overlap_border * 2:
                            over_tok_ = over_tok + overlap_border
                                # change starts here
                            cut_tok_ = cut_tok - overlap_border
                                # change ends here (not including)
                            overlap = cut_tok_ - over_tok_
                                # length of changeable area
                            sent_over_tok_ = overlap_border
                                # start of the eff part of the sent
                            sent_cut_tok_ = sent_over_tok_ + overlap
                                # end of the eff part of the sent
                            step = 1 / (overlap + 1)
                            token_starts_ = token_starts[orig_sent_idx]
                            over_sub_ = token_starts_[over_tok_]
                            cut_sub_ = token_starts_[cut_tok_]
                            weights = (
                                torch.tensor(sub_to_kens[orig_sent_idx]
                                                        [over_sub_:cut_sub_],
                                             device=data_device)
                              - over_tok_ + 1
                            ) * step
                            weights.unsqueeze_(1)
                            sent_token_starts_ = token_starts[sent_idx]
                            sent_token0_start_ = sent_token_starts_[0]
                            sent_over_sub_ = \
                                1 + sent_token_starts_[sent_over_tok_] \
                                  - sent_token0_start_
                            sent_cut_sub_ = \
                                1 + sent_token_starts_[sent_cut_tok_] \
                                  - sent_token0_start_
                            # weighs are used to merge the overlap zone
                            overlap_data = \
                                orig_data[over_sub_:cut_sub_] * (1
                                                               - weights) \
                              + sent[sent_over_sub_:sent_cut_sub_] * weights

                        else:
                            over_sub_ = overlap_border
                            sent_cut_sub_ = 1 + over_sub_
                            overlap_data = None

                        data[orig_sent_idx] = torch.cat((
                            orig_data[:over_sub_],
                            *([] if overlap_data is None else [overlap_data]),
                            sent[sent_cut_sub_:sent_fin]
                        ), dim=0)

                    else:
                        data.append(sent[1:sent_fin])

        ## sort data in original order
        ######
        if need_sort_dataset:
            _, data = zip(*sorted(
                [(i, x) for i, x in enumerate(data)],
                key=lambda x: sorted_sent_ids[x[0]]
            ))
            data = list(data)
        ######

        if aggregate_subtokens_op != 'append':
            _src = num_subtokens
            if loglevel:
                print('Reordering')
                _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)
            for i, token_lens in enumerate(_src):
                #token_lens = num_subtokens[i]
                start = 0
                sent = data[i]
                sent_ = []
                for token_len in token_lens:
                    end = start + token_len
                    vecs = sent[start:end]
                    if aggregate_subtokens_op != None:
                        vecs = self._aggregate_hidden_states(
                            vecs, layer_ids=None,
                            aggregate_op=aggregate_subtokens_op
                        )
                    sent_.append(vecs)
                    start = end

                data[i] = sent_ if aggregate_subtokens_op == None else \
                          torch.stack(sent_, dim=0)

        if save:
            if append:
                self.data += data
            else:
                self.data = data
        else:
            return data

    def _collate(self, batch, with_lens=True, with_token_lens=True,
                 append_subtokens=False):
        """The method to use with `torch.utils.data.DataLoader` and
        `.transform_collate()`.

        :with_lens: return lengths of data.
        :with_token_lens: return lengths of tokens of the data (only allowed
            if `.transform()` was called with `aggregate_subtokens_op=None`).
        :append_subtokens: line up tokens if `.transform()` was called with
            `aggregate_subtokens_op=None`. The result will be exactly as if
            `aggregate_subtokens_op='append'` but you can get token lens.
        :return: depends on keyword args.
        :rtype: if the `.transform()` method was called with
            *aggregate_subtokens_op*=None:
                tuple(list([torch.tensor]), lens:torch.tensor,
                      token_lens:list([torch.tensor]))
            otherwise: tuple(list([torch.tensor]), lens:torch.tensor)
        """
        device = CPU
        pad = 0.

        if isinstance(batch[0], Tensor):
            if batch[0].is_cuda:
                device = batch[0].get_device()
            lens = [tensor([len(x) for x in batch], device=device,
                           dtype=self.int_tensor_dtype)] if with_lens else []
            x = pad_sequences_with_tensor(batch, padding_tensor=pad)

        else:
            for x in batch:
                if x:
                    if x[0].is_cuda:
                        device = x[0].get_device()
                    tensor_dtype = x[0].dtype
                    break
            lens = [tensor([len([x for x in x for x in x]) for x in batch],
                           device=device, dtype=self.int_tensor_dtype)] \
                       if with_lens and append_subtokens else \
                   [tensor([len(x) for x in batch], device=device,
                           dtype=self.int_tensor_dtype)] \
                       if with_lens else \
                   []
            if with_token_lens:
                lens.append([tensor([len(x) for x in x], device=device,
                                    dtype=self.int_tensor_dtype)
                                 for x in batch])
            if append_subtokens:
                x = pad_sequences_with_tensor(
                    [torch.vstack([x for x in x for x in x]) for x in batch],
                    padding_tensor=pad
                )
            else:
                x = pad_array_torch(batch, padding_value=pad,
                                    device=device, dtype=tensor_dtype)

        return (x, *lens)



# -*- coding: utf-8 -*-
# junky lib: utils
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a bunch of utilities to use with PyTorch.
"""
from asyncio import Lock as ALock
from collections.abc import Iterable
from collections import OrderedDict, Counter
from copy import deepcopy
from itertools import chain
from tqdm import tqdm
import numpy as np
import random
from threading import Lock as TLock
import torch
import math

CPU = torch.device('cpu')


def save_model(model: torch.nn.Module, path: str):
    """Saves the **model** with zero gradients and moved to the CPU."""
    torch.save(model, path)
    model = torch.load(path, map_location=CPU)
    torch.optim.Optimizer(model.parameters(), {}).zero_grad()
    model.eval()
    torch.save(model, path)

def load_model(path: str, device=None):
    """Just a wrapper for `torch.load()`."""
    return torch.load(path, map_location=device)

def get_max_dims(array, str_isarray=False, max_dims=None, dim_no=0):
    """Returns max sizes of nested *array* on the all levels of nestedness.

    :param array: nested lists or tuples.
    :param str_isarray: if True, strings are treated as arrays of chars and
        form additional dimension.
    :param max_dims: for internal use only. Leave it as it is.
    :param dim_no: for internal use only. Leave it as it is.
    """
    if max_dims is None:
        max_dims = []  # it's importaint! don't try to move
                       # max_dims=[] to the args list!

    max_dims_len = len(max_dims)
    res = max_dims
    isnew = dim_no == len(max_dims)
    if not isnew:
        max_dim = max_dims[dim_no]
        if max_dim is None:
            res = None

    if res is not None:
        array_len = None
        isstr = isinstance(array, str)
        try:
            if not str_isarray and isstr:
                raise TypeError()
            array_len = len(array)
        except TypeError:
            res = None

        if isnew:
            max_dims.append(array_len)
        elif array_len is None:
            max_dims[dim_no:] = [array_len]
        else:
            if array_len > max_dim:
                max_dims[dim_no] = array_len
            if isstr:
                max_dims[dim_no + 1:] = [None]

        if res is not None and not isstr:
            dim_no_ = dim_no + 1
            for el in array:
                if not get_max_dims(el, str_isarray=str_isarray,
                                    max_dims=max_dims, dim_no=dim_no_):
                    break

    if dim_no == 0 and res is not None:
        res = res[:-1]

    return res

def insert_to_ndarray(array, ndarray, shift='left'):
    """Inserts a nested *array* with data of any allowed for numpy type to the
    numpy *ndarray*. NB: all dimensions of *ndarray* must be no less than sizes
    of corresponding subarrays of the *array*.

    :param array: nested lists or tuples.
    :param shift: how to place data of *array* to *ndarray* in the case if
        size of some subarray less than corresponding dimension of *ndarray*.
        Allowed values:
            'left': shift to start;
            'right': shift to end;
            'center': place by center (or 1 position left from center if
                      evennesses of subarray's size and ndarray dimension are
                      not congruent);
            'rcenter': the same as 'center', but if evennesses are not
                       congruent, the shift will be 1 position right.
    """
    asize = len(array)
    nsize = len(ndarray)

    if shift == 'left':
        start_idx = 0
    elif shift == 'right':
        start_idx = nsize - asize
    elif shift in ['center', 'lcenter']:
        start_idx = (nsize - asize) // 2
    elif shift == 'rcenter':
        start_idx = int((nsize - asize + 1) // 2)
    else:
        raise ValueError('ERROR: Unknown shift value "{}"'.format(shift))

    if len(ndarray.shape) > 1:
        for idx in range(len(array)):
            insert_to_ndarray(array[idx], ndarray[start_idx + idx],
                              shift=shift)
    else:
        ndarray[start_idx:start_idx + len(array)] = array

def pad_array(array, padding_value=0):
    """Converts nested *array* with data of any allowed for numpy type to
    numpy.ndarray with *padding_value* instead of missing data.

    :param array: nested lists or tuples.
    :rtype: numpy.ndarray
    """

    dims = get_max_dims(array)
    out_array = np.full(shape=dims, fill_value=padding_value)
    insert_to_ndarray(array, out_array)

    return out_array

def pad_array_torch(array, padding_value=0, **kwargs):
    """Just a wropper for ``pad_array()`` that returns *torch.Tensor*.

    :param kwargs: keyword args for the ``torch.tensor()`` method.
    :rtype: torch.Tensor
    """
    return torch.tensor(pad_array(array, padding_value=padding_value),
                        **kwargs)

def pad_sequences_with_tensor(sequences, padding_tensor=0.):
    """Pads the `seq` dimension of *sequences* of shape
    `(batch, seq, features)` with *padding_tensor*.

    :param sequences: the list of sequences of shape `(seq, features)`.
    :type sequences: list([torch.Tensor(seq, features)]]
    :param padding_tensor: scalar or torch.Tensor of the shape of `features`
        to pad *sequences*.
    :type padding_tensor: torch.Tensor(features)|float|int
    :return: padded list converted to torch.Tensor.
    :rtype: torch.Tensor(batch, seq, features)
    """
    t = sequences[0]
    device = t.get_device() if t.is_cuda else CPU
    N, S = len(sequences), max(s.shape[0] for s in sequences)
    if not isinstance(padding_tensor, torch.Tensor):
        if isinstance(padding_tensor, Iterable):
            padding_tensor = torch.tensor(padding_tensor, device=device)
        else:
            padding_tensor = t.new_full(t.shape[1:], padding_tensor)
    #res = padding_tensor.to(device).expand(N, S, *t.shape[2:]).clone()
    res = padding_tensor.to(device).repeat(N, S, 1)

    for s_i, s in enumerate(sequences):
        res[s_i, :s.shape[0]] = s

    return res

def enforce_reproducibility(seed=None):
    """Re-inits random number generators.
    [We stole this method from Stanford C224U assignments]"""
    if seed:
        # Sets seed manually for both CPU and CUDA
        torch.manual_seed(seed)
        # For atomic operations there is currently
        # no simple way to enforce determinism, as
        # the order of parallel operations is not known.
        #
        # CUDNN
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        # System based
        np.random.seed(seed)
        random.seed(seed)

def get_rand_vector(shape, norm, shift=0., dtype=np.float64):
    """Creates random vector with the norm given.

    :param shape: the shape of the new vector.
    :type shape: tuple|list
    :param norm: the norm of the new vector.
    :type norm: float
    :param shift: relative shift of the new vector's mean from 0.
    :type shift: float
    :rtype: numpy.ndarray
    """
    if norm:
        vector = np.random.rand(*shape).astype(dtype) - .5 + shift
        vector *= norm / np.linalg.norm(vector)
    else:
        vector = np.zeros(shape)
    return vector

def add_mean_vector(vectors, axis=0, shift=0., scale=1.):
    """Appends *vectors* with a vector that has norm equals to the mean norm
    (possibly, scaled) of *vectors*.

    :param vectors: the array of floats.
    :type vectors: numpy.ndarray
    :param axis: the axis of *vectors* along which the new vector is appended.
    :type axis: int|tuple(int)
    :param shift: relative shift of the new vector's mean from 0.
    :type shift: float
    :param scale: the coef to increase (or decrease if *scale* < 1) the norm
        or the new vector.
    :type scale: float > 0
    :return: vectors appended
    :rtype: numpy.ndarray
    """
    norm = np.linalg.norm(vectors.shape[:axis] + vectors.shape[axis + 1:],
                          vectors, axis=axis).mean() * scale
    norm = np.linalg.norm(vectors, axis=tuple(
        list(range(0)) + list(range(1, len(emb_model.vectors.shape)))
    )).mean()
    vector = np.expand_dims(get_rand_vector(norm, shift, dtype=vectors.dtype))
    return np.append(vectors, vector, axis=axis)

def absmax(array, axis=None):
    """Returns an array with absolute maximum values of *array* according to
    *axis*.

    :type array: numpy.ndarray|list([numpy.ndarray])
    :type axis: int
    :rtype: numpy.ndarray
    """
    if not isinstance(array, np.ndarray):
        array = list(array)
        assert len(array) > 0, 'ERROR: array has no data'
        assert isinstance(array[0], np.ndarray), \
               'ERROR: array has unsupported type'
        array = np.stack(array, axis=-1 if axis is None else axis)

    if axis is None:
        res = array.max()
        res_min = array.min()
        if -res_min > res:
            res = res_min

    else:
        amax = np.broadcast_to(
            np.expand_dims(
                abs(array).argmax(axis=axis), axis=axis
            ), array.shape
        )

        shape = [(*[None] * len(amax.shape))]
        shape[axis] = ...
        mask = np.broadcast_to(
            np.arange(amax.shape[axis])[tuple(shape)],
            array.shape
        )

        res = array.copy()
        res[mask != amax] = 0

        res = res.sum(axis=axis)

    return res

def absmax_torch(tensors, dim=None):
    """Returns a tensor with absolute maximum values of *tensors* according to
    *dim*.

    :type tensors: torch.Tensor|list([torch.Tensor])
    :type dim: int
    :rtype: torch.Tensor
    """
    if not isinstance(tensors, torch.Tensor):
        tensors = list(tensors)
        assert len(tensors) > 0, 'ERROR: tensors has no data'
        assert isinstance(tensors[0], torch.Tensor), \
               'ERROR: tensors have unsupported type'
        tensors = torch.stack(tensors, dim=-1 if dim is None else dim)

    if dim is None:
        res = tensors.max()
        res_min = tensors.min()
        if -res_min > res:
            res = res_min

    else:
        amax = tensors.abs() \
                      .argmax(dim=dim) \
                      .unsqueeze(dim=dim) \
                      .expand(tensors.shape)

        shape = [(*[None] * len(amax.shape))]
        shape[dim] = ...
        mask = torch.arange(amax.shape[dim])[tuple(shape)] \
                    .expand(amax.shape)

        if tensors.is_cuda:
            mask = mask.to(tensors.device)

        res = tensors.clone()
        res[mask != amax] = 0

        res = res.sum(dim=dim)

    return res

def kwargs(**kwargs):
    """Returns any keyword arguments as a dict."""
    return kwargs

def kwargs_nonempty(**kwargs):
    """Returns any keyword arguments with non-empty values as a dict."""
    return {x: y for x, y in kwargs.items() if y}

def get_func_params(func, func_locals, keep_self=False):
    """Returns params of *func* as `args` and `kwargs` arrays. Method is
    called inside *func*; *func_locals* is an output of the locals() call
    inside *func*.

    If *keep_self* is ``True``, don't remove `self` variable from `args`."""
    all_args = func.__code__.co_varnames[:func.__code__.co_argcount]
    n_kwargs = len(func.__defaults__)
    args = [func_locals[x] for x in all_args[:-n_kwargs]
                if x != 'self' or keep_self]
    kwargs = {x: func_locals[x] for x in all_args[-n_kwargs:]}
    return args, kwargs

def add_class_lock(cls, lock_name='lock', isasync=False, lock_object=None):
    """Adds additional lock property *lock_name* to class *cls*.

    if *isasync* is ``True`` the lock property is of `asyncio.Lock` type.
    Otherwise (default) it's of `threading.Lock` type.

    It can be used as follows:

    from junky add_class_lock
    from pkg import Cls
    add_class_lock(Cls)

    o = Cls()
    #async with o.lock:  # if isasync is True
    with o.lock:         # if isasync is False (default)
        # some thread safe operations here
        pass

    Also, you can add lock to the particular object directly:

    o = add_class_lock(Cls())

    If you need, you can use your own lock object. Use param *lock_object* for
    that. In that case, param *isasync* is ignored.
    """
    if not lock_object:
        lock_object = ALock() if isasync else TLock()
    '''
    _code = cls.__init__.__code__
    co_varnames, co_argcount = _code.co_varnames, _code.co_argcount
    #co_kwonlyargcount = _code.co_kwonlyargcount
    _defaults = cls.__init__.__defaults__
    #_kwdefaults = cls.__init__.__kwdefaults__
    Cls = type('Cls', (cls,),
               {x: _defaults[i] for i, x in \
                    enumerate(co_varnames[co_argcount - len(_defaults) \
                             :co_argcount])})
    setattr(Cls, lock_name, property(lambda self: lock))
    return Cls
    '''
    setattr(cls, lock_name,
            property(lambda self: lock_object) if isinstance(cls, type) else
                     lock_object)
    return cls

def filter_embeddings(pretrained_embs, corpus, min_abs_freq=1, save_name=None,
                   include_emb_info=False, pad_token=None, unk_token=None,
                   extra_tokens=None):
    """Filters pretrained word embeddings' vocabulary, leaving only tokens
    that are present in the specified `corpus` which are more frequent than
    minimum absolute frequency `min_abs_freq`. This method allows to
    significantly reduce memory usage and speed up word embedding process.
    The drawbacks include lower performance on unseen data.

    Args:

    **vectors**: file with pretrained word vectors in text format (not
    binary), where the first line is
    `<vocab_size> <embedding_dimensionality>`.

    **corpus**: a list of lists or tuples with already tokenized sentences.
    Filtered result will not contain any tokens outside of this corpus.

    **min_abs_freq** (`int`): minimum absolute frequency; only tokens the
    frequency of which is equal or greater than this specified value will be
    included in the filtered word embeddings. Default `min_abs_freq=1`,
    meaning all words from the corpus that have corresponding word vectors in
    `pretrained_embs` are preserved.

    **save_name**(`str`): if specified, filtered word embeddings are saved in
    a file with the specified name.

    **include_emb_info**(`bool`): whether to include `<vocab_size> <emb_dim>`
    as the first line to the filtered embeddings file. Default is `False`,
    embedding info line is skipped. Relevant only if `save_name` is not None.

    For the arguments below note, that typically pretrained embeddings already
    include PAD or UNK tokens. But these argumets are helpful if you need to
    specify your custom pad/unk/extra tokens or make sure they are at the top
    of the vocab (thus, pad_token will have index=0 for convenience).

    **pad_token** (`str`): custom padding token, which is initialized with
    zeros and included at the top of the vocabulary.

    **unk_token** (`str`): custom token for unknown words, which is
    initialized with small random numbers and included at the top of the
    vocabulary.

    **extra_tokens** (`list`): list of any extra custom tokens. For now, they
    are initialized with small random numbers and included at the top of the
    vocabulary. Typically, used for special tokens, e.g. start/end tokens etc.

    If `save_name` is specified, saves the filtered vocabulary. Otherwise,
    returns word2index OrderedDict and a numpy array of corresponding word
    vectors.
    """

    filter_vocab = OrderedDict(
        sorted(
            {k: v
             for k, v in Counter(chain.from_iterable(corpus)).items()
             if v>=min_abs_freq}.items(),
            key=lambda t: t[1]))

    word2index = OrderedDict()
    vectors = []

    # model in vec or txt format
    # (not binary, first line is <vocab_size> <emb_dim>)
    word2vec_file = open(pretrained_embs)

    n_words, embedding_dim = word2vec_file.readline().split()
    n_words, embedding_dim = int(n_words), int(embedding_dim)

    if pad_token:
        # Zero vector for PAD
        vectors.append(np.zeros((1, embedding_dim)))
        word2index[pad_token] = len(word2index)

    if unk_token:
        # Initializing UNK vector with small random numbers
        vectors.append(
            np.random.rand(1, embedding_dim) / math.sqrt(embedding_dim))
        word2index[unk_token] = len(word2index)

    if extra_tokens:
        # random-small-number vectors for extra_tokens
        for x_t in extra_tokens:
            vectors.append(
                np.random.rand(1, embedding_dim) / math.sqrt(embedding_dim))
            word2index[x_t] = len(word2index)

    progress_bar = tqdm(desc='Filtering vectors',
                        total=n_words, mininterval=2)

    while True:
        line = word2vec_file.readline().strip()

        if not line:
            break

        current_parts = line.split()
        current_word = ' '.join(current_parts[:-embedding_dim])

        if current_word in filter_vocab:

            word2index[current_word] = len(word2index)

            current_vectors = current_parts[-embedding_dim:]
            current_vectors = np.array(list(map(float, current_vectors)))
            current_vectors = np.expand_dims(current_vectors, 0)

            vectors.append(current_vectors)

        progress_bar.update(1)

    progress_bar.close()
    word2vec_file.close()

    vectors = list(np.concatenate(vectors))

    if save_name:
        with open(save_name, 'w') as f:
            if include_emb_info:
                print(len(word2index), embedding_dim, file=f)

            for word, vector in tqdm(zip(word2index.keys(), vectors),
                         desc='Saving filtered vectors', total=len(vectors)):
                print(word, ' '.join(str(v) for v in vector),
                      end=' \n', file=f)
    else:
        return word2index, vectors

def balance_positive_values(data, distinction_coef=1.5, attractor='middle'):
    """Divides an array of positive values into groups, such as sums of their
    elements are close to each other.

    Args:

    **data**: an array of positive (greater than `0`) numbers.

    **distinction_coef**: the maximal coefficient of distinction between the
    maximum and minimum **data** values belonging to the same group.

    **attractor**: the value used to calculate the number of groups. Possible
    values are:<br />
    `'max'` - *\<maximum data value>*<br />
    `'lower'` - `'max'` / **distinction_coef**<br />
    `'min'` - the value from **data**` immediately preceding `'lower'`<br />
    `'upper'` - *`'min'` * **distinction_coef**<br />
    `'mean'` - the middle point between `'upper'` and `'lower'`<br />
    `'middle'` (default) - the middle point between `'max'` and `'min'`<br />
    In order to calculate the number of groups, the sum of all the values of
    **data** is divided to that value. Also, it is possibly just to specify a
    desired sum of the elements for each group as a value for **attractor**.
    The `'+'`/`'-'` sign before text **attractor** value increases/decreases the
    resulting number of groups by `1`.

    Returns a `list` of groups found. If a group contain just a single number,
    the corresponding element of the returning `list` is just that number.
    Otherwise (if group contain multiple numbers), the element is an array of
    numbers.
    """
    vals = sorted(data, reverse=True)
    num_folds_add = 0
    if isinstance(attractor, str):
        max_val = vals[0]
        lower_bound = max_val / distinction_coef
        min_ix = None
        for ix, val in enumerate(vals):
            if val < lower_bound:
                min_ix, min_val = ix, vals[ix - 1]
                break
        if isinstance(attractor, str):
            if attractor[-1] == '-':
                num_folds_add = 1
                attractor = attractor[:-1]
            elif attractor[-1] == '+':
                num_folds_add = -1
                attractor = attractor[:-1]
    else:
        min_ix = -1
    if min_ix and min_ix < len(vals) - 1:
        if isinstance(attractor, str):
            # vals, tail = [[x] for x in vals[:min_ix]], vals[min_ix:]
            vals, tail = vals[:min_ix], vals[min_ix:]
            upper_bound = min_val * distinction_coef
        else:
            vals, tail = [], vals
        sum_tail = sum(tail)
        num_folds = int(sum_tail // (
            upper_bound if attractor == 'upper' else
            lower_bound if attractor == 'lower' else
            max_val if attractor == 'max' else
            min_val if attractor == 'min' else
            lower_bound + (upper_bound - lower_bound) / 2
                if attractor == 'mean' else
            min_val + (max_val - min_val) / 2
                if attractor == 'middle' else
            attractor
        )) + num_folds_add or 1
        fold_sums, tail = tail[:num_folds], tail[num_folds:]
        folds = [[x] for x in fold_sums]
        while tail:
            for ix, val in zip(reversed(range(num_folds)), tail):
                folds[ix].append(val)
                fold_sums[ix] += val
            tail = tail[num_folds:]
            folds.sort(reverse=True)
        vals += folds
    # else:
    #     vals = [[x] for x in vals[:min_ix]]
    return vals

def balance_intents(y_data, distinction_coef=1.5, attractor='middle',
                    tail_prefix='Tail', inplace=False):
    """Divides an array of labels into groups, such as counts of their
    elements are close to each other.

    Args:

    **y_data**: an array of labels.

    **distinction_coef** and **attractor**: see help for
    `balance_positive_values()` function.

    **tail_prefix**: new composite groups (ones that contain not identical
    elements of **y_data**) will get names of that value trailed with integer
    number.

    Returns 3 arrays:

    *classes balanced*: a `list` of new target categories. If a category
    contain just a single label from the original set, the corresponding
    element of the returning `list` is just that label. Otherwise (if a
    category contain multiple original labels), the element is a `tuple`,
    the 1st element of which is a name of that new category (prefixed with
    **tail_prefix**), and the 2nd one is the `list` of original labels that
    it contains.

    *class lengths balanced*: the counts of elements from **y_data** which
    labels are contained in the corresponding elements of *classes balances*.
    Here, for single categories, their lengths are just numbers, whereas for
    the composite categories their corresponding lengths are `list` of
    separate lengths for each included original label.

    *new y_data*: **y_data** with values replaced by names of new target
    categories.
    """
    assert y_data, 'ERROR: y_data must be not empty.'
    ismulti =  isinstance(y_data[0], Iterable)
    y_data_ = [x for x in y_data for x in x] if ismulti else y_data
    assert y_data_, 'ERROR: y_data must be not empty.'
    classes = {}
    for cls in y_data_:
        classes[cls] = classes.get(cls, 0) + 1
    classes, class_lens = zip(*classes.items())
    classes, class_lens = list(classes), list(class_lens)
    class_lens_balanced = balance_values(class_lens,
                                         distinction_coef=distinction_coef,
                                         attractor=attractor)
    classes_balanced, classes_tail = [], {}
    tail_ix = -1
    for class_len in class_lens_balanced:
        if len(class_len) > 1:
            tail_ix += 1
            class_name = tail_prefix + str(tail_ix)
            class_tail = []
            classes_balanced.append((class_name, class_tail))
            for class_len in class_len:
                ix = class_lens.index(class_len)
                class_ = classes.pop(ix)
                class_tail.append(class_)
                class_lens.pop(ix)                
                classes_tail[class_] = class_name
        else:
            ix = class_lens.index(class_len)
            classes_balanced.append(classes.pop(ix))
            class_lens.pop(ix)
    if not inplace:
        y_data = deepcopy(y_data)
    for ix, class_ in enumerate(y_data):
        if ismulti:
            for ix, cls in enumerate(class_):
                new_class = classes_tail.get(cls)
                if new_class is not None:
                    class_[ix] = new_class
        else:
            new_class = classes_tail.get(class_)
            if new_class is not None:
                y_data[ix] = new_class
    return classes_balanced, class_lens_balanced, y_data

def rotate_vectors(vecs, cosine, d_cosine=0, norm_coef=1, d_norm_coef=0,
                   dim=-1):
    """Rotates vectors of **vecs** in random directions to the angle defined
    with **cosine**.

    Args:

    **vecs**: either a `list`, `numpy.ndarray` or `torch.Tensor` that we want
    to rotate. Any dimensionality is allowed.

    **cosine** (`float`): the cosine of the angle on which the **vecs** should
    be rotated.

    **d_cosine** (`float`): if specified, the real **cosine** will be randomly
    selected from the range `[cosine, cosine + d_cosine)`.

    **norm_coef** (`float`): the L2 norm of the resulting vectors will be set
    as `L2_norm(vecs) * norm_coef`.

    **d_norm_coef** (`float`): if specified, the real **norm_coef** will be
    randomly selected from the range `[norm_coef, norm_coef + d_norm_coef)`.

    **dim** (`int`): the rotation will be made relative to that dimension.

    Returns rotated and resized copy of **vecs**."""
    convert_to = 'numpy' if isinstance(vecs, np.ndarray) else \
                 'list' if isinstance(vecs, list) else \
                 None
    if convert_to:
        vecs = torch.tensor(vecs)
    elif not isinstance(vecs, torch.Tensor):
        raise ValueError(
            'ERROR: **vecs** must be of either `torch.Tensor`, '
           f'`numpy.ndarray` or `list` type. Found `{type(vecs)}` type '
            'instead.'
        )

    shape_ = list(vecs.shape)
    f_norm = torch.norm

    # L2 norms of input vecs
    vecs_norm = f_norm(vecs, dim=dim)
    # creating random vectors to define rotation planes for each vector of
    # `vecs`
    hypos = torch.rand(*shape_)
    # next, we're gonna construct a set of rectangular triangles for which
    # `vecs` would be cathetuses, `hypos` after resize would be hypotenuses,
    # and the set of another cathetuses we still need to find
    cos = torch.nn.functional.cosine_similarity(vecs, hypos, dim=dim)
        # cosines between vecs and hypos
    hypos_norm = vecs_norm / cos  # lengths of hypos to be hypotenuses
    hypos *= (hypos_norm / f_norm(hypos, dim=dim)).unsqueeze(dim=dim)
        # now `hypos` are real hypotenuses
    cats = hypos - vecs
        # 2nd cathetuses

    shape_.pop(dim)
    if d_cosine != 0:
        cosine += d_cosine * torch.rand(*shape_, device=vecs.device)
    if d_norm_coef != 0:
        norm_coef += d_norm_coef * torch.rand(*shape_, device=vecs.device)

    # the cossines of angles to rotate are given, so we can resize 2nd
    # cathetuses which opposed to those angles. then, hypothenuses of such
    # triangles will be onw-way with our target vectors
    cats_norm = vecs_norm * np.sqrt(1 / (cosine * cosine) - 1)
        # required lengths for the 2nd cathetuses
    cats *= (cats_norm / f_norm(cats, dim=dim)).unsqueeze(dim=dim)
        # cathetus are resized
    res = vecs + cats
        # hypothenuses
    # resizing hypothenuses to target norm values, we'll obtain the target set
    # of vectors
    res *= (vecs_norm * norm_coef / f_norm(res, dim=dim)).unsqueeze(dim=dim)

    return res.numpy() if convert_to == 'numpy' else \
           res.tolist() if convert_to == 'list' else \
           res
